/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/


#include "stdafx.h"
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"


#define theWnd ((CMainWindow)dc.theWindow)


////// The item edit object //////


CDrawEditItem::CDrawEditItem( CTinyCadDoc *pDesign )
: CDrawingObject( pDesign )
{
  segment=1;
  EditMethodText = -1;
  InMove = FALSE;
  InSelectByDrag = FALSE;
  LastPos = CPoint(-1,-1);
}

void CDrawEditItem::EndEdit()
{
  if (!segment)
  {
	(m_pDesign->GetHead())->EndEdit();
  }

  m_pDesign->UnSelect();
  m_pDesign->Invalidate();

  segment=1;
}


void CDrawEditItem::NewOptions()
{
	CDrawingObject *theEditItem = m_pDesign->GetSelect();

	if (theEditItem != NULL)
		theEditItem->NewOptions();

}

void CDrawEditItem::Move(CPoint p)
{
  CPoint r = CPoint(p.x - LastPos.x,p.y - LastPos.y);

  if (EditMethodText!=-1) 
  {
	  m_pDesign->SetSelectable( NULL );
	m_pDesign->GetHead()->MoveField(EditMethodText,r);
  } 
  else if (InSelectByDrag) 
  {
	  m_pDesign->SetSelectable( NULL );
  	Display();
  	b = p;
  	Display();
  } 
  else if (InMove) 
  {
	  m_pDesign->SetSelectable( NULL );
  	m_pDesign->SelectMove(r);
  }
  else
  {
	// Is there anything under the pointer?
	CDrawingObject *pointer = m_pDesign->GetHead();
	while (pointer != NULL) 
	{
		if (pointer->IsInside(p.x,p.x,p.y,p.y))
		{
			break;
		}
		pointer = pointer->next;
	}
	m_pDesign->SetSelectable( pointer );

  }

  LastPos = p;	
}


int CDrawEditItem::SetCursor( CPoint p )
{
	if (m_pDesign->GetSelect() != NULL) 
	{

	   	// Is this object currently being edited a method object?
	  	if (m_pDesign->GetHead()==m_pDesign->GetSelect()) {
			// Has the user clicked on one of the method fields?
			return m_pDesign->GetHead()->SetCursorEdit( p );
  		}
		else
		{
			CDrawingObject *pointer = m_pDesign->GetHead();
			while (pointer != NULL) {
				if (pointer->IsInside(p.x,p.x,p.y,p.y))
					return 8;
				if (pointer == m_pDesign->GetSelect())
					break;
				pointer = pointer->next;
			}
		}
	}

	return -1;
}

void CDrawEditItem::Place(CPoint p, CPoint no_snap_p)
{
	LastPos = p;

	// Where we moving a method field?
   	if (EditMethodText != -1)
		return;

	// Have some items already been selected, and are we clicking on one?
	if (m_pDesign->GetSelect() != NULL) 
	{
		a = b = p;

	   	// Is this object currently being edited a method object?
	  	if (m_pDesign->GetHead()==m_pDesign->GetSelect()) 
		{
			// Has the user clicked on one of the method fields?
			EditMethodText = m_pDesign->GetHead()->IsInsideField(no_snap_p);
			if (EditMethodText!=-1)
				return;
  		}

		// No, so see if any of the other objects can be moved...
		CDrawingObject *pointer = m_pDesign->GetHead();
		while (pointer != NULL) 
		{
			if (pointer->IsInside(no_snap_p.x,no_snap_p.x,no_snap_p.y,no_snap_p.y))
				InMove = TRUE;
			if (pointer == m_pDesign->GetSelect())
				break;
			pointer = pointer->next;
		}
	} 
	else 
	{
		// Start the select box...
		a = p;
		b = p;

		InMove = FALSE;
 		InSelectByDrag = TRUE;

	  	Display();
	}
}


void CDrawEditItem::Paint(Context &dc,paint_options options)
{

  dc.SelectBrush();
  dc.SelectPen(PS_DOT,1,cBLOCK);
  dc.SetROP2(R2_COPYPEN);

  dc.Rectangle(CRect(a.x,a.y,b.x,b.y));
}



int CDrawEditItem::CanUseOptions()
{
  // What is the object currently being edited?
  if (!segment)
  	return (m_pDesign->GetHead())->CanUseOptions();
  else
  	return cdoDefault;

}


void CDrawEditItem::EndPlace(CPoint p)
{

  // Keep a copy of this position
  LastPos = p;

  // Are we in a move?
  if (InMove) 
  {
  	InMove = FALSE;
	
	// If we actually moved the objects then must create an undo path
	if (p != a) 
	{
		m_pDesign->SelectUndoMove(CPoint(a.x-p.x,a.y-p.y));
		m_pDesign->IncUndoLevel();
	}

	m_pDesign->Invalidate();
	
	return;
  }

  // Turn off the drag box (if there is one...)
  if (InSelectByDrag) 
  {
  	Display();
	InSelectByDrag = FALSE;
  }
  else
  {
  	a = b = CPoint(0,0);
  }

  // Were we moving a method field, if so stop and return
  if (EditMethodText != -1) 
  {
	EditMethodText = -1;
	m_pDesign->Invalidate();
	return;
  }


  // *** The main processing part....
  // If this was just a single click then select one item
  // If this was a dragged box then select all the items inside the box!
  if (a == b) 
  {
  	// Unselect the edited items
  	End();

  	// Find which item is under the pointer
  	CDrawingObject *pointer=m_pDesign->GetHead(),*last;

  	if (pointer==NULL)
		return;

  	// Search each object in turn, leaving the object at the head until last
  	last=pointer;
  	pointer=pointer->next;
  	while (pointer!=NULL) {

		if (pointer->IsInside(p.x,p.x,p.y,p.y) )
			break;

		// Move pointer on
		last=pointer;
		if (pointer==m_pDesign->GetHead())
			return;
		pointer=pointer->next;
		if (pointer==NULL)
			pointer=m_pDesign->GetHead();
  	}
	
	if (pointer!=NULL && pointer!=m_pDesign->GetHead()) {

		// Move selected item to the head
		m_pDesign->Select(pointer);

		// Now insert the head object at the tail
		CDrawingObject *head = pointer->next;
		if (head != NULL) 
		{
			CDrawingObject *look = pointer;
			while (look!=NULL) 
			{
				last=look;
				look=look->next;
			}
			// Check to see if the head is at the tail...
			if (head != last)
			{
				pointer->next = head->next;
				last->next=head;
				head->next=NULL;
			}
		}
  	}

  	// Can the selected object be edited?
  	if ( m_pDesign->GetHead()==NULL )
	{
		m_pDesign->Invalidate();
		return;
	}

	// Now create a copy of this selected object
  	CDrawingObject *head = m_pDesign->GetHead();
  	head->Store();

  	// Swap the new object and the old object
  	head = m_pDesign->GetHead();
  	pointer = head->next;
  	head->next = pointer->next;
  	pointer->next = head;
  	m_pDesign->SetHead(pointer);

  	// Now select the original object to delete it (for the Undo list)
  	m_pDesign->SetSelect(m_pDesign->GetHead());
  	m_pDesign->SelectDelete();
	m_pDesign->IncUndoLevel();

  	// Select this item
  	(m_pDesign->GetHead())->Display();
  	m_pDesign->SetSelect(m_pDesign->GetHead());
	  } else {
  	//*** Must select a block of items
	//*** Select so they can be moved or cut/copied

  	// Unselect the edited items
  	End();

	// Select the items in the box (if any)
	m_pDesign->Select(a,b);
	a = b = CPoint(0,0);
  }


  // If only one item selected then start editing it!
  if (m_pDesign->GetHead()!= NULL
  	&& m_pDesign->GetSelect() == m_pDesign->GetHead() 
  	&& (m_pDesign->GetHead())->CanEdit()) {
  	// Begin Editing this Item
  	(m_pDesign->GetHead())->BeginEdit();

  	segment=0;
  } else 
  	segment=1;

//  m_pDesign->Invalidate();	
}

void CDrawEditItem::ReleaseSelection()
{
  // If in the middle of a move then cancel it!
  InMove = FALSE;

  // If necessary remove the edit item dialog
  if (!segment && m_pDesign->GetHead()!=NULL && m_pDesign->GetHead() == m_pDesign->GetSelect()) {
	(m_pDesign->GetHead())->EndEdit();
	(m_pDesign->GetHead())->Display();
  }

  segment=1;

}


void CDrawEditItem::AfterEnd(CPoint p, CPoint)
{
	// Select a new object for the menu...

  	// Find which item is under the pointer
  	CDrawingObject *pointer=m_pDesign->GetHead(),*last;

  	if (pointer==NULL)
		return;

  	// Search each object in turn, leaving the object at the head until last
  	last=pointer;
  	pointer=pointer->next;
  	while (pointer!=NULL) {

		if (pointer->IsInside(p.x,p.x,p.y,p.y)  )
			break;

		// Move pointer on
		last=pointer;
		if (pointer==m_pDesign->GetHead())
			return;
		pointer=pointer->next;
		if (pointer==NULL)
			pointer=m_pDesign->GetHead();
  	}
	
	if (pointer!=NULL && pointer!=m_pDesign->GetHead()) {

		// Move selected item to the head
		m_pDesign->Select(pointer);

		// Now insert the head object at the tail
		CDrawingObject *head = pointer->next;
		if (head != NULL) {
			CDrawingObject *look = pointer;
			while (look!=NULL) {
				last=look;
				look=look->next;
			}
			pointer->next = head->next;
			last->next=head;
			head->next=NULL;
		}
  	}

  	// Is the selected object a method object?
  	if ( m_pDesign->GetHead()==NULL || m_pDesign->GetHead()->GetType() != xMethodEx2)
		return;

  	// Now create a copy of this selected object
  	CDrawingObject *head = m_pDesign->GetHead();
  	head->Store();

  	// Swap the new object and the old object
  	head = m_pDesign->GetHead();
  	pointer = head->next;
  	head->next = pointer->next;
  	pointer->next = head;
  	m_pDesign->SetHead(pointer);

  	// Now select the original object to delete it (for the Undo list)
  	m_pDesign->SetSelect(m_pDesign->GetHead());
  	m_pDesign->SelectDelete();
	m_pDesign->IncUndoLevel();

  	// Select this item
  	(m_pDesign->GetHead())->Display();
  	m_pDesign->SetSelect(m_pDesign->GetHead());

	// Now unselect the object
	End();
}

BOOL CDrawEditItem::End()
{
  // If in the middle of a move then cancel it!
  if (InMove)
  {
  	m_pDesign->SelectMove(CPoint(a.x - LastPos.x, a.y - LastPos.y));
	InMove = FALSE;
  }

  // If necessary remove the line from the screen
  if (!segment && m_pDesign->GetHead()!=NULL && m_pDesign->GetHead() == m_pDesign->GetSelect()) 
  {
	(m_pDesign->GetHead())->EndEdit();
	(m_pDesign->GetHead())->Display();
  }


  // Unselect the last object
  m_pDesign->UnSelect();

  segment=1;

  return TRUE;
}

