/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/



// This handles the actual drawing of objects
#include "stdafx.h"
#include <stdlib.h>
#include "TinyCadView.h"
#include "diag.h"
#include "colour.h"
#include "option.h"


#define theWnd ((CMainWindow)dc.theWindow)

// The Power Class //

#define POWER_SIZE	14



void CDrawPower::TagResources()
{
 	m_pDesign->GetOptions()->TagFont(fPIN);
}


int DoRotate(int olddir,int newdir)
{
  //     New rotation=> 2  3  4  
  //				    Current dir ..\/..
  int table[] = {	2, 3, 0, 		// 0 (Up)
			3, 2, 1,		// 1 (Down)
			1, 0, 3,		// 2 (Left)
			0, 1, 2,		// 3 (Right)
		};

  return table[(newdir-2) + olddir*3];
}

void CDrawPower::Load(CArchive &archive)
{
 	a = ReadPoint(archive);
  archive >> which >> dir >> str;
  b=a;
  segment=0;
}

void CDrawPower::Save(CArchive &archive)
{
  WritePoint(archive,a);
  archive << which << dir << str;
}


int CDrawPower::IsInside(int left,int right,int top,int bottom)
{
  return !( (TextPos.x<left && TextEnd.x<=left) || (TextPos.x>right  && TextEnd.x>=right)
         || (TextPos.y<top  && TextEnd.y<=top)  || (TextPos.y>bottom && TextEnd.y>=bottom) )

       || (a.x>=left && a.x<=right && a.y>=top && a.y<=bottom);
}

// Look for a seach string in the object
CString CDrawPower::Find(const char *theSearchString)
{
  CString HoldString;

  HoldString = str;
  HoldString.MakeLower();

  // Now look for the search string in this string
  return (HoldString.Find(theSearchString) != -1) ? str : "";
}


// Rotate this object about a point
void CDrawPower::Rotate(CPoint p,int ndir)
{			 
  // Translate this point so the rotational point is the origin
  a = CPoint(a.x-p.x,a.y-p.y);
  b = CPoint(b.x-p.x,b.y-p.y);

  // Perfrom the rotation
  switch (ndir) {
	case 2: // Left
		a = CPoint(a.y,-a.x);
		b = CPoint(b.y,-b.x);
		break;		
	case 3: // Right
		a = CPoint(-a.y,a.x);
		b = CPoint(-b.y,b.x);
		break;
	case 4: // Mirror
		a = CPoint(-a.x,a.y);
		b = CPoint(-b.x,b.y);
		break;
  }

  // Re-translate the points back to the original location
  a = CPoint(a.x+p.x,a.y+p.y);
  b = CPoint(b.x+p.x,b.y+p.y);

  dir = DoRotate(dir,ndir);
}


BOOL CDrawPower::CanEdit()
{
  return TRUE;
}

ObjType CDrawPower::GetType()
{
  return xPower;
}

CString CDrawPower::GetName()
{
  return "Power";
}


CDrawPower::CDrawPower(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
  segment=1;
  a=CPoint(0,0);
  b=a;
  str="";
  dir=1;
  which=0;

  Undo_Level=0; 
  Delete_Level=-1;
}

void CDrawPower::BeginEdit()
{
  g_EditToolBar.m_PowerEdit.Open(m_pDesign,this);
}

void CDrawPower::Move(CPoint p)
{
  Display();
  a=p;
  b=a;
  segment=0;
  Display();
}

void CDrawPower::Place(CPoint p, CPoint)
{
  Display();
  a=p;
  Store();
  Display();	// Write to screen
}


void CDrawPower::Display( BOOL erase )
{
	// Invalidate the symbol
	CRect r( a.x,a.y,b.x,b.y);
	m_pDesign->InvalidateRect( r, erase, POWER_SIZE * 2 + 2 );

	r.left = TextPos.x;
	r.top = TextPos.y;
	r.right = TextEnd.x;
	r.bottom = TextEnd.y;
	m_pDesign->InvalidateRect( r, erase, 5 );
}


// Display the power item on the screen!
void CDrawPower::Paint(Context &dc,paint_options options)
{
  int spacing;
  CSize size;
  CPoint pa,pb,pc,pd,pos=a;

  dc.SelectFont(*m_pDesign->GetOptions()->GetFont(fPIN),2);
  dc.SetROP2(R2_COPYPEN);
  dc.SetTextColor(cPOWER);

  size=dc.GetTextExtent(str);
  if (which!=0)
	spacing=POWER_SIZE*2+POWER_SIZE/4;
  else
	spacing=POWER_SIZE+POWER_SIZE/4;

  // Find out which way round the power object goes
  switch (dir) {
	case 0:		// Top
		pa=CPoint(pos.x,pos.y-POWER_SIZE);
		pb=CPoint(pos.x-POWER_SIZE/2,pos.y-POWER_SIZE);
		pc=CPoint(pos.x+POWER_SIZE/2+1,pos.y-POWER_SIZE);
		pd=CPoint(pos.x,pos.y-POWER_SIZE*2);
		TextPos = CPoint(pos.x-size.cx/2,pos.y-spacing);
		break;
	case 1:		// Bottom
		pa=CPoint(pos.x,pos.y+POWER_SIZE);
		pb=CPoint(pos.x-POWER_SIZE/2,pos.y+POWER_SIZE);
		pc=CPoint(pos.x+POWER_SIZE/2+1,pos.y+POWER_SIZE);
		pd=CPoint(pos.x,pos.y+POWER_SIZE*2);
		TextPos = CPoint(pos.x-size.cx/2,pos.y+spacing+size.cy);
		break;
	case 2:		// Left
		pa=CPoint(pos.x-POWER_SIZE,pos.y);
		pb=CPoint(pos.x-POWER_SIZE,pos.y-POWER_SIZE/2);
		pc=CPoint(pos.x-POWER_SIZE,pos.y+POWER_SIZE/2+1);
		pd=CPoint(pos.x-POWER_SIZE*2,pos.y);
		TextPos=CPoint(pos.x-size.cx-spacing,pos.y+size.cy/2);
		break;
	case 3:		// Right
		pa=CPoint(pos.x+POWER_SIZE,pos.y);
		pb=CPoint(pos.x+POWER_SIZE,pos.y-POWER_SIZE/2);
		pc=CPoint(pos.x+POWER_SIZE,pos.y+POWER_SIZE/2+1);
		pd=CPoint(pos.x+POWER_SIZE*2,pos.y);
		TextPos=CPoint(pos.x+spacing,pos.y+size.cy/2);
		break;
  }

  dc.TextOut(str,TextPos,options);
  TextEnd = CPoint(TextPos.x + size.cx,TextPos.y - size.cy);

  switch (options)
  {
  case draw_selected:
  case draw_selectable:
	  dc.SelectPen(PS_SOLID,1,cSELECT);
	  break;
  case draw_handles:
	  break;
  default:
	  dc.SelectPen(PS_SOLID,1,cLINE);
  }

  
  dc.SelectBrush();


  // Draw the main line of the power item
  switch (which) {
	case 0: // Draw the Bar
  		dc.MoveTo(pb);
  		dc.LineTo(pc);
		break;
	case 1: // Draw the Circle
		dc.Ellipse(CRect(pb.x-(pa.x-pd.x),pb.y-(pa.y-pd.y),pc.x,pc.y));
		break;
	case 2: // Draw the Wave
		// The PolyBezier function is not implemented in MFC!
		// So this hack has to be used....
		{
		  POINT pts[4];
		  pts[0]=CPoint(pb.x+(pd.x-pa.x)/2,pb.y+(pd.y-pa.y)/2);
		  pts[1]=CPoint(pd.x+(pb.x-pa.x)/2,pd.y+(pb.y-pa.y)/2);
		  pts[2]=CPoint(pa.x+(pc.x-pa.x)/2,pa.y+(pc.y-pa.y)/2);
		  pts[3]=CPoint(pc.x-(pa.x-pd.x)/2,pc.y-(pa.y-pd.y)/2);
		  dc.PolyBezier(pts,4);
		}
		break;
	case 3: // Draw the Arrow
		dc.MoveTo(pb);
  		dc.LineTo(pc);
		dc.MoveTo(pb);
		dc.LineTo(pd);
		dc.LineTo(pc);
		break;
	case 4: // Draw the Earth
  		dc.MoveTo(pb);
  		dc.LineTo(pc);
		dc.MoveTo(CPoint(pb.x-(pb.x-pd.x)/2,pb.y-(pb.y-pd.y)/2));
		dc.LineTo(CPoint(pc.x-(pc.x-pd.x)/2,pc.y-(pc.y-pd.y)/2));
		dc.MoveTo(CPoint(pd.x-(pb.x-pa.x)/4,pd.y-(pb.y-pa.y)/4));
		dc.LineTo(CPoint(pd.x+(pb.x-pa.x)/4,pd.y+(pb.y-pa.y)/4));
		break;
  }



  // Draw the bits that all power items have in comman
  dc.MoveTo(pos);
  dc.LineTo(pa);
}


void CDrawPower::EndEdit()
{
  End();
  g_EditToolBar.m_PowerEdit.Close();
}

// Store the line in the drawing
void CDrawPower::Store()
{
  CDrawPower *NewObject;

  NewObject = new CDrawPower(m_pDesign);
  *NewObject = *this;

  m_pDesign->Add(NewObject);

  m_pDesign->SetRepeat(NULL);
}





////// The Pin Class //////


void CDrawPin::TagResources()
{
 	m_pDesign->GetOptions()->TagFont(fPIN);
}


void CDrawPin::OldLoad(CArchive &archive)
{

  a = ReadPoint(archive);
  archive >> which >> elec >> dir >> part >> str >> number >> show;
  b=a;
  segment=0;

  // Sneekly hide the smallpin parameter in the which parameter....
  BOOL smallpin = which & 128;
  which &= ~128;

  length = (smallpin ? 20 : 30);
}

void CDrawPin::Load(CArchive &archive)
{
  a = ReadPoint(archive);
  archive >> which >> elec >> dir >> part >> str >> number >> show >> length;
  b=a;
  segment=0;

  if (length < 5)
  {
	  length = 20;
  }
}

void CDrawPin::Save(CArchive &archive)
{
  WritePoint(archive,a);
  archive << which << elec << dir << part << str << number << show << length;
}


int CDrawPin::IsInside(int left,int right,int top,int bottom)
{
  // Only inside if it is in the current part
  return IsInvisible() ? FALSE : (a.x>=left && a.x<=right && a.y>=top && a.y<=bottom);
}

BOOL CDrawPin::CanEdit()
{
  return TRUE;
}

ObjType CDrawPin::GetType()
{
  return xPinEx;
}


CString CDrawPin::GetName()
{
  return "Pin";
}

// Look for a seach string in the object
CString CDrawPin::Find(const char *theSearchString)
{
  CString HoldString;

  HoldString = str;
  HoldString.MakeLower();

  // Now look for the search string in this string
  return (HoldString.Find(theSearchString) != -1) ? str : "";
}


// Construct a pin
CDrawPin::CDrawPin(CTinyCadDoc *pDesign)
: CDrawingObject( pDesign )
{
  segment=1;
  which=0;
  dir=0;
  number="1";
  elec=0;
  str="";
  show=3;
  length = pDesign->GetOptions()->GetPinLength();
  part = m_pDesign->GetPart();
  Undo_Level=0;
  Delete_Level=-1;
}


CPoint CDrawPin::GetPosition(CDrawMethod *parent)
{
  CPoint r,p;
  CPoint bound = parent->GetTr();

  // Has this object been mirrored?
  if ((parent->GetRotate()&4)!=0)
	p = CPoint(-a.x + bound.x ,a.y);
  else
	p = a;

  switch ((parent->GetRotate())&3) {
	case 0:	// Up
		r.x = p.x + parent->a.x;
		r.y = p.y + parent->a.y;
		break;
	case 1: // Down
		r.x = p.x + parent->a.x;
		r.y = (bound.y - p.y) + parent->a.y;
		break;
	case 2: // Left
		r.x = p.y + parent->a.x;
		r.y = p.x + parent->a.y;
		break;
	case 3: // Right
		r.x = (bound.y - p.y) + parent->a.x;
		r.y = p.x + parent->a.y;
		break;
  }


  return r;
}


void CDrawPin::BeginEdit()
{
  g_EditToolBar.m_PinEdit.Open(m_pDesign,this);
}

void CDrawPin::Move(CPoint p)
{
  Display();
  a=p;
  b=a;
  segment=0;
  Display();
}



// Determine weather or not this pin is visible
int CDrawPin::IsInvisible()
{
	int i = m_pDesign->GetPart();
 bool r = (part != i || (IsPower()  && !(m_pDesign->IsEditLibrary())) );

 return r;
}


#define PIN_UNSCALED	length
#define PIN_SIZE	PIN_UNSCALED

#define SIZING		((PIN_SIZE <= 20) ? 20 : 30)
#define DOT_SIZE	(SIZING/6)
#define LINE_SIZE	(PIN_SIZE-DOT_SIZE*2)
#define TEXT_SPACE	(SIZING/8)

void CDrawPin::DetermineLayout( CPoint &pa,CPoint &pb,CPoint &pc,CPoint &pd,CPoint &pta,CPoint &ptb, int &dr, int rotmir )
{
  int spacing;
  CPoint pos = a;
  CSize size;

  dr=3;

  if ((which&2)==0)
	spacing=PIN_SIZE+TEXT_SPACE;
  else
	spacing=PIN_SIZE+(DOT_SIZE*2)+TEXT_SPACE;

  // Find out which way round the pin object goes
  switch (dir) {
	case 0:		// Up
		pa=CPoint(pos.x,pos.y-LINE_SIZE);
		pb=CPoint(pos.x-DOT_SIZE,pos.y-PIN_SIZE);
		pc=CPoint(pos.x+DOT_SIZE+1,pos.y-PIN_SIZE);
		pd=CPoint(pos.x,pos.y-PIN_SIZE);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x+size.cy/2,pos.y-spacing);
		ptb=CPoint(pos.x,pos.y-TEXT_SPACE);
		dr=0;
		break;
	case 1:		// Down
		pa=CPoint(pos.x,pos.y+LINE_SIZE);
		pb=CPoint(pos.x-DOT_SIZE,pos.y+PIN_SIZE);
		pc=CPoint(pos.x+DOT_SIZE+1,pos.y+PIN_SIZE);
		pd=CPoint(pos.x,pos.y+PIN_SIZE);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x+size.cy/2,pos.y+spacing+size.cx);
		size=m_pDesign->GetTextExtent(number, fPIN);
		ptb=CPoint(pos.x,pos.y+size.cx+TEXT_SPACE);
		dr=0;
		break;
	case 2:		// Left
		pa=CPoint(pos.x-LINE_SIZE,pos.y);
		pb=CPoint(pos.x-PIN_SIZE,pos.y-DOT_SIZE);
		pc=CPoint(pos.x-PIN_SIZE,pos.y+DOT_SIZE+1);
		pd=CPoint(pos.x-PIN_SIZE,pos.y);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x-size.cx-spacing,pos.y+size.cy/2);
		size=m_pDesign->GetTextExtent(number, fPIN);
		ptb=CPoint(pos.x-size.cx-TEXT_SPACE,pos.y);
		break;
	case 3:		// Right
		pa=CPoint(pos.x+LINE_SIZE,pos.y);
		pb=CPoint(pos.x+PIN_SIZE,pos.y-DOT_SIZE);
		pc=CPoint(pos.x+PIN_SIZE,pos.y+DOT_SIZE+1);
		pd=CPoint(pos.x+PIN_SIZE,pos.y);

		size=m_pDesign->GetTextExtent(str, fPIN);
		pta=CPoint(pos.x+spacing,pos.y+size.cy/2);
		ptb=CPoint(pos.x+TEXT_SPACE,pos.y);
		break;
  }



  // Make corrections for bizzar rotations
  switch (dir)
  {
  case 2:
  case 3:
	    switch (rotmir)
		{
		case 1: // Down
		case 5: // Down & Mirror
		case 7: // Right
		case 3: // Right & Mirror
		  pta=CPoint(pta.x, pta.y - size.cy );
		  ptb=CPoint(ptb.x, ptb.y - size.cy );
		  break;
		}
  }

}


void CDrawPin::Display( BOOL erase )
{

  CPoint pa,pb,pc,pd,pta,ptb;
  int dr;
  DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, 0 );

	// Invalidate the symbol
	CRect r( a.x -2,a.y, pd.x + 2,pd.y);
	m_pDesign->InvalidateRect( r, erase, DOT_SIZE + 6 );


  if ((show&1)!=0)
  {
	  CSize sz = m_pDesign->GetTextExtent( str, fPIN );
	  if (dr < 2)
	  {
		  r = CRect( pta.x - sz.cy, pta.y - sz.cy, pta.x + sz.cy, pta.y + sz.cx);
	  }
	  else
	  {
		  r = CRect( pta.x, pta.y - sz.cy, pta.x + sz.cx, pta.y + sz.cy );
	  }
	  m_pDesign->InvalidateRect( r, erase, 2 );
  }
  if ((show&2)!=0)
  {
  	  CSize sz = m_pDesign->GetTextExtent( number, fPIN );
	  if (dr < 2)
	  {
		  r = CRect( ptb.x - sz.cy, ptb.y - sz.cy, ptb.x + sz.cy, ptb.y + sz.cx);		  
	  }
	  else
	  {
		  r = CRect( ptb.x, ptb.y - sz.cy, ptb.x + sz.cx, ptb.y + sz.cy );
	  }
	  
	  m_pDesign->InvalidateRect( r, erase, 2 );	
  }

}


// Display the pin item on the screen!
void CDrawPin::Paint(Context &dc,paint_options options)
{
  // Only display the pin if the it is in the current part
  if (IsInvisible() || (which == 5 && !(m_pDesign->IsEditLibrary())) )
	return;

  CPoint pa,pb,pc,pd,pta,ptb;
  int dr;
  DetermineLayout( pa,pb,pc,pd,pta,ptb,dr, dc.theTransform.GetRotMir() );

  dc.SetROP2(R2_COPYPEN);
  dc.SetTextColor(cPOWER);

  dc.SelectBrush();
  dc.SelectFont(*m_pDesign->GetOptions()->GetFont(fPIN),dr);

  switch (options)
  {
  case draw_selected:
  case draw_handles:
	    dc.SelectPen(PS_SOLID,1,cSELECT );
	  break;
  case draw_selectable:
	  dc.SelectPen(PS_SOLID,1,cPIN_CLK);
	  break;
  default:
	    dc.SelectPen(PS_SOLID,1,cPIN);
  }


  if ((which & 1)!=0 && which < 4 ) 
  {
	// Draw the dot
	dc.Ellipse(CRect(pb.x+(pa.x-pd.x),pb.y+(pa.y-pd.y),pc.x,pc.y));
  } 
  else 
  {
	dc.MoveTo(pa);
	dc.LineTo(pd);
  }

  // Draw the clock (triangle)
  if ((which & 2)!=0) 
  {
	dc.MoveTo(pb);
  	dc.LineTo(pc);
	dc.MoveTo(pb);
	dc.LineTo(CPoint(pd.x-(pa.x-pd.x),pd.y-(pa.y-pd.y)));
	dc.LineTo(pc);
  }



  // Draw the bits that all pins have in comman
  dc.MoveTo(a);
  dc.LineTo(pa);

  if ((show&1)!=0)
	dc.TextOut(str,pta,options,dr);
  if ((show&2)!=0)
	dc.TextOut(number,ptb,options,dr);

}


void CDrawPin::EndEdit()
{
  End();
  g_EditToolBar.m_PinEdit.Close();
}


void CDrawPin::Place(CPoint p, CPoint)
{
  Display();
  a=p;
  b=a;
  Store();
  Display();	// Write to screen

  // Add 1 to the pin number (if possible)
  int hold=atoi(number);
  // Only increment if old value was a number
  if (hold!=0) {
	hold += m_pDesign->GetPinDir();
	if (hold>999 || hold<1)
		hold=1;
	_itoa(hold++,number.GetBuffer(17),10);
    number.ReleaseBuffer();
  }

  hold=str.FindOneOf("0123456789-");
  // Only increment if old value was a number
  if (hold!=-1) {
	char NewNum[17];
	_itoa(atoi(str.Mid(hold))+m_pDesign->GetNameDir(),NewNum,10);
	str = str.Left(hold)+NewNum;
  }


  g_EditToolBar.m_PinEdit.ReFocus();

  Display();
}



// Rotate this object about a point
void CDrawPin::Rotate(CPoint p,int ndir)
{
  // Translate this point so the rotational point is the origin
  a = CPoint(a.x-p.x,a.y-p.y);
  b = CPoint(b.x-p.x,b.y-p.y);

  // Perfrom the rotation
  switch (ndir) {
	case 2: // Left
		a = CPoint(a.y,-a.x);
		b = CPoint(b.y,-b.x);
		break;		
	case 3: // Right
		a = CPoint(-a.y,a.x);
		b = CPoint(-b.y,b.x);
		break;
	case 4: // Mirror
		a = CPoint(-a.x,a.y);
		b = CPoint(-b.x,b.y);
		break;
  }

  // Re-translate the points back to the original location
  a = CPoint(a.x+p.x,a.y+p.y);
  b = CPoint(b.x+p.x,b.y+p.y);

  dir = DoRotate(dir,ndir);
}



// Store the line in the drawing
void CDrawPin::Store()
{
  CDrawPin *NewObject;

  NewObject = new CDrawPin(m_pDesign);
  *NewObject = *this;

  m_pDesign->Add(NewObject);
  m_pDesign->SetRepeat(NULL);
}

