/*
	TinyCAD program for schematic capture
	Copyright 1994/1995/2002 Matt Pyne.

	This program is free software; you can redistribute it and/or
	modify it under the terms of the GNU Lesser General Public
	License as published by the Free Software Foundation; either
	version 2.1 of the License, or (at your option) any later version.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
	Lesser General Public License for more details.

	You should have received a copy of the GNU Lesser General Public
	License along with this library; if not, write to the Free Software
	Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/

#include "stdafx.h"
#include "TinyCadView.h"


////// The Symbol Field class //////

// Save this symbol field into an archive
void CSymbolField::Save(CArchive &ar)
{
	DWORD ft = field_type;
	ar << field_name << field_value << field_default << field_pos << ft;
}

// Load this symbol field from an archive
void CSymbolField::Load(CArchive &ar)
{
	DWORD ft;
	ar >> field_name >> field_value >> field_default >> field_pos >> ft;
	field_type = (SymbolFieldType)ft;
}



////// The Symbol Record class //////

CDrawingObject *CSymbolRecord::LoadSymbol( CTinyCadDoc *pDesign, CArchive &ar, CPoint &tr )
{
  // Move to the right location in filespace
  CDrawingObject *pSelect = NULL;
  try
  {
	  pSelect = pDesign->GetSelect();

	  // Load in the design
	  pDesign->Load(ar);
  }
  catch (...)
  {
	  // Failed to load the symbol
	  return NULL;
  }

  // Convert the design into a linked list of a method
  CDrawingObject *method=pDesign->GetHead();

  // Remove this from the design list
  pDesign->DetachSelect();
  pDesign->SetSelect(pSelect);

  // Find the co-ords of the bounding box of this symbol
  CDrawingObject *pointer=method;

  CPoint a=CPoint(0,0);
  CPoint b=pointer->a;

  while (pointer!=NULL) {
	a.x=max(a.x,max(pointer->a.x,pointer->b.x));
	a.y=max(a.y,max(pointer->a.y,pointer->b.y));

	b.x=min(b.x,min(pointer->a.x,pointer->b.x));
	b.y=min(b.y,min(pointer->a.y,pointer->b.y));
	
	pointer=pointer->next;
  }

  // Snap the co-ords of the bounding box to the grid
  CPoint snapa=pDesign->m_snap.Snap(a);
  CPoint snapb=pDesign->m_snap.Snap(b);
  if (snapa.x<a.x)
	snapa.x+=pDesign->m_snap.GetGrid();
  if (snapa.y<a.y)
	snapa.y+=pDesign->m_snap.GetGrid();
  if (snapb.x>b.x)
	snapb.x-=pDesign->m_snap.GetGrid();
  if (snapb.y>b.y)
	snapb.y-=pDesign->m_snap.GetGrid();


  // Now translate so that bottom left hand corner is at 0,0
  pointer=method;

  while (pointer!=NULL) {
	pointer->a=CPoint(pointer->a.x-snapa.x,pointer->a.y-snapa.y);
	pointer->b=CPoint(pointer->b.x-snapa.x,pointer->b.y-snapa.y);
	pointer=pointer->next;
  }

  // Store top right hand corner
  tr=CPoint(snapb.x-snapa.x,snapb.y-snapa.y);

  return method;
}




////// The Symbol class //////

// The constructor
CLibraryFileSymbol::CLibraryFileSymbol(CLibraryFile *NewLib)
{
  lib=NewLib;
}

// Another constructor...
CLibraryFileSymbol::CLibraryFileSymbol()
{
  lib = NULL;
}

// The destructor
CLibraryFileSymbol::~CLibraryFileSymbol()
{
}

CString CLibraryFileSymbol::GetLibName()
{
  return lib->name;
}



// Is this symbol correct filled in?
BOOL CLibraryFileSymbol::IsValid()
{
	return name!="" && reference!="" && ppp>0 && (name.Find(' ') == -1);
}


// Does this string match this symbol
CString CLibraryFileSymbol::Find(const char *theSearchString)
{
  CString HoldString;

  HoldString = name + "  " + description;
  HoldString.MakeLower();

  // Now look for the search string in this string
  return (HoldString.Find(theSearchString) != -1) ? name + " " + description : "";
}


// Save this symbol into an archive
void CLibraryFileSymbol::Save(CArchive &theArchive)
{
  DWORD number_of_fields = fields.size();
  theArchive << name << description << reference << ppp << number_of_fields; 

  for (int i = 0; i < number_of_fields; i++)
  {
	  fields[i].Save( theArchive );
  }
}


// load this symbol from an archive
void CLibraryFileSymbol::Load(CArchive &theArchive)
{
  DWORD number_of_fields;
  FilePos = 0;
  theArchive >> name >> description >> reference >> ppp >> number_of_fields; 
  for (int i = 0; i < number_of_fields; i++)
  {
	  CSymbolField sf;
	  sf.Load( theArchive );
	  fields.push_back( sf );
  }
}

// load this symbol from an archive
void CLibraryFileSymbol::OldLoad3(CArchive &theArchive)
{
	CString dummy;
  theArchive >> name >> description >> reference >> ppp >> FilePos >> dummy; 
}


// load this symbol from an archive
void CLibraryFileSymbol::OldLoad1(CArchive &theArchive)
{
  theArchive >> name >> description >> reference >> ppp >> FilePos; 
}


// the old version of load this symbol from an archive
void CLibraryFileSymbol::OldLoad2(CArchive &theArchive)
{
  theArchive >> name >> reference >> ppp >> FilePos; 
  description = "";
}



// Get the object this symbol refers to, this method, may not be in memory
// and so will have to be extracted from the library file.
CDrawingObject *CLibraryFileSymbol::GetMethod( CTinyCadDoc *pDesign, CPoint &tr )
{
  // No, so must load it from the methods file

  CFile *theMethods=lib->GetMethodFile();
  
  if (!theMethods)
	return NULL;

  try
  {
	  theMethods->Seek(FilePos,CFile::begin);
  }
  catch (...)
  {
	  // Failed to load the symbol
	  return NULL;
  }

  CArchive ar( theMethods, CArchive::load );
  return LoadSymbol( pDesign, ar, tr );
}



CDesignFileSymbol *CLibraryFileSymbol::GetDesignSymbol( CTinyCadDoc *pDesign )
{
	CDrawingObject *pMethod;
	CPoint tr;
	pMethod = GetMethod( pDesign, tr );
	CDesignFileSymbol *psymbol = new CDesignFileSymbol( pDesign, pMethod, tr );

	psymbol->description = description;
	psymbol->name = name;
	psymbol->ppp = ppp;
	psymbol->reference = reference;

	return psymbol;
}



///// The symbol as contained by the design /////

CDesignFileSymbol::CDesignFileSymbol(CTinyCadDoc *pDesign, CDrawingObject *pMethod, CPoint tr )
{
	m_pDesign = pDesign;
	m_pMethod = pMethod;
	m_Tr = tr;
}

// The Destructor
CDesignFileSymbol::~CDesignFileSymbol()
{
	CDrawingObject *pointer = m_pMethod;

	while (pointer != NULL) {
		CDrawingObject *next = pointer->next;
		delete pointer;
		pointer = next;
	}
}
	
// Save this symbol into an archive
void CDesignFileSymbol::Save(CArchive &theArchive)
{
    DWORD number_of_fields = fields.size();
	CString dummy;
	dummy.Format("%d", number_of_fields );
	theArchive << name << description << reference << ppp << dummy; 
	for (int i = 0; i < number_of_fields; i++)
	{
		fields[i].Save( theArchive );
	}


	CDrawingObject *oldHead = m_pDesign->GetHead();
	m_pDesign->SetHead( m_pMethod );
	m_pDesign->Save(theArchive,FALSE,FALSE);
	m_pDesign->SetHead( oldHead );
}

// Load this symbol from an archive
void CDesignFileSymbol::Load(CTinyCadDoc *pDesign, CArchive &theArchive)
{
	CString dummy;
	theArchive >> name >> description >> reference >> ppp >> dummy; 

	DWORD number_of_fields = 0;
	if (!dummy.IsEmpty())
	{
		number_of_fields = atoi( dummy );
	}

	for (int i = 0; i < number_of_fields; i++)
	{
	  CSymbolField sf;
	  sf.Load( theArchive );
	  fields.push_back( sf );
	}

	m_pDesign = pDesign;
	m_pMethod = LoadSymbol( pDesign, theArchive, m_Tr );
}


// Return the methods object
CDrawingObject *CDesignFileSymbol::GetMethod( CPoint &tr )
{
	tr = m_Tr;
	return m_pMethod;
}
